"""
!pip install tensorflow-addons
!pip install -q -U tensorflow-text
!pip install -q -U tf-models-official
!pip install -U tfds-nightly

"""

import os
import tensorflow as tf
import tensorflow_hub as hub
import tensorflow_text as text  # A dependency of the preprocessing model
# import tensorflow_addons as tfa  # For metrics
# from official.nlp import optimization

tf.get_logger().setLevel('ERROR')
"""
Configure TFHub to read checkpoints directly from TFHub's Cloud Storage buckets. This is only recommended when running TFHub models on TPU.
Without this setting TFHub would download the compressed file and extract the checkpoint locally. Attempting to load from these local files will fail with the following error:
```
InvalidArgumentError: Unimplemented: File system scheme '[local]' not implemented
```
This is because the [TPU can only read directly from Cloud Storage buckets](https://cloud.google.com/tpu/docs/troubleshooting#cannot_use_local_filesystem).
Note: This setting is automatic in Colab.
"""
os.environ["TFHUB_MODEL_LOAD_FORMAT"] = "UNCOMPRESSED"


class BERT_FineTune:
    def __init__(self, handle_encoder, handle_preprocess):
        if tf.config.list_physical_devices('TPU'):
            cluster_resolver = tf.distribute.cluster_resolver.TPUClusterResolver(tpu='')
            tf.config.experimental_connect_to_cluster(cluster_resolver)
            tf.tpu.experimental.initialize_tpu_system(cluster_resolver)
            strategy = tf.distribute.TPUStrategy(cluster_resolver)
            print('Using TPU')
        elif tf.config.list_physical_devices('GPU'):
            strategy = tf.distribute.MirroredStrategy()
            print('Using GPU')
        else:
            strategy = None
            print('Running on CPU is not recommended.')
        self.strategy = strategy
        self.handle_encoder = handle_encoder
        self.handle_preprocess = handle_preprocess

    @staticmethod
    def make_bert_preprocess_model(handle_preprocess, sentence_features, seq_length=128):
        """Returns Model mapping string features to BERT inputs.

        Args:
        sentence_features: a list with the names of string-valued features.
        seq_length: an integer that defines the sequence length of BERT inputs.

        Returns:
        A Keras Model that can be called on a list or dict of string Tensors
        (with the order or names, resp., given by sentence_features) and
        returns a dict of tensors for input to BERT.
        """

        input_segments = [
            tf.keras.layers.Input(shape=(), dtype=tf.string, name=ft) for ft in sentence_features
        ]

        # Tokenize the text to word pieces.
        bert_preprocess = hub.load(handle_preprocess)
        tokenizer = hub.KerasLayer(bert_preprocess.tokenize, name='tokenizer')
        segments = [tokenizer(s) for s in input_segments]

        # Optional: Trim segments in a smart way to fit seq_length.
        # Simple cases (like this example) can skip this step and let
        # the next step apply a default truncation to approximately equal lengths.
        truncated_segments = segments

        # Pack inputs. The details (start/end token ids, dict of output tensors)
        # are model-dependent, so this gets loaded from the SavedModel.
        packer = hub.KerasLayer(
            bert_preprocess.bert_pack_inputs, arguments=dict(seq_length=seq_length), name='packer'
        )
        model_inputs = packer(truncated_segments)
        return tf.keras.Model(input_segments, model_inputs)

    @staticmethod
    def convert_dataset(
        df,
        batch_size,
        bert_preprocess_model,
        sentence_features,
        label='label',
        shuffle=False,
        repeat=False
    ):
        AUTOTUNE = tf.data.AUTOTUNE
        in_memory_ds = dict()
        for feature in sentence_features:
            in_memory_ds[feature] = df[feature]
        in_memory_ds['label'] = df[label]

        dataset = tf.data.Dataset.from_tensor_slices(in_memory_ds)
        num_examples = len(in_memory_ds['label'])

        if shuffle:
            dataset = dataset.shuffle(num_examples)
        if repeat:
            dataset = dataset.repeat()
        dataset = dataset.batch(batch_size)
        dataset = dataset.map(lambda ex: (bert_preprocess_model(ex), ex['label']))
        dataset = dataset.cache().prefetch(buffer_size=AUTOTUNE)
        return dataset, num_examples

    @staticmethod
    def build_classifier_model(encoder, num_classes, seed=0):
        tf.random.set_seed(seed)

        class Classifier(tf.keras.Model):
            def __init__(self, encoder, num_classes):
                super(Classifier, self).__init__(name="prediction")
                self.encoder = encoder
                self.dropout = tf.keras.layers.Dropout(0.1)
                self.dense = tf.keras.layers.Dense(num_classes)

            def call(self, preprocessed_text):
                encoder_outputs = self.encoder(preprocessed_text)
                pooled_output = encoder_outputs["pooled_output"]
                x = self.dropout(pooled_output)
                x = self.dense(x)
                return x

        model = Classifier(encoder, num_classes)
        return model

    def tune(
        self,
        train_df,
        val_df=None,
        sentence_features=['sentence'],
        label='label',
        epochs=3,
        batch_size=32,
        optimizer='sgd',
        seq_length=128
    ):
        num_classes = len(train_df[label].unique())
        bert_preprocess_model = BERT_FineTune.make_bert_preprocess_model(
            self.handle_preprocess, sentence_features, seq_length
        )

        train_dataset, train_data_size = BERT_FineTune.convert_dataset(
            train_df, batch_size, bert_preprocess_model, sentence_features, label, True, True
        )

        steps_per_epoch = train_data_size // batch_size
        num_train_steps = steps_per_epoch * epochs
        num_warmup_steps = num_train_steps // 10

        if val_df is None:
            validation_dataset = None
            validation_steps = None
        else:
            validation_dataset, validation_data_size = BERT_FineTune.convert_dataset(
                val_df, batch_size, bert_preprocess_model, sentence_features, label, False, False
            )
            validation_steps = validation_data_size // batch_size

        if self.strategy is None:
            encoder = hub.KerasLayer(self.handle_encoder, trainable=True)

            # metric have to be created inside the strategy scope

            loss = tf.keras.losses.SparseCategoricalCrossentropy(from_logits=True)

            # metrics = tfa.metrics.MatthewsCorrelationCoefficient(num_classes=2)
            classifier_model = BERT_FineTune.build_classifier_model(encoder, num_classes)

            # optimizer = optimization.create_optimizer(
            #     init_lr=init_lr,
            #     num_train_steps=num_train_steps,
            #     num_warmup_steps=num_warmup_steps,
            #     optimizer_type='adamw')

            classifier_model.compile(
                optimizer=optimizer, loss=loss, metrics='accuracy'
            )  # metrics=[metrics]

            classifier_model.fit(
                x=train_dataset,
                validation_data=validation_dataset,
                steps_per_epoch=steps_per_epoch,
                epochs=epochs,
                validation_steps=validation_steps
            )
        else:
            with self.strategy.scope():
                encoder = hub.KerasLayer(self.handle_encoder, trainable=True)

                # metric have to be created inside the strategy scope

                loss = tf.keras.losses.SparseCategoricalCrossentropy(from_logits=True)

                # metrics = tfa.metrics.MatthewsCorrelationCoefficient(num_classes=2)

                classifier_model = BERT_FineTune.build_classifier_model(encoder, num_classes)

                # optimizer = optimization.create_optimizer(
                #     init_lr=init_lr,
                #     num_train_steps=num_train_steps,
                #     num_warmup_steps=num_warmup_steps,
                #     optimizer_type='adamw')

                classifier_model.compile(
                    optimizer=optimizer, loss=loss, metrics='accuracy'
                )  # metrics=[metrics]

                classifier_model.fit(
                    x=train_dataset,
                    validation_data=validation_dataset,
                    steps_per_epoch=steps_per_epoch,
                    epochs=epochs,
                    validation_steps=validation_steps
                )
        return classifier_model



from collections import Counter
import random


def up_resample(x_train, y_train, reach_percent=.5, seed=0):
    """
    function will copy paste labels with less data,up to the reach_percent level of biggest label in set
    :param x_train:
    :param y_train:
    :param reach_percent: float, reach at least what percentage of largest set in data,
    fo labels already equal or higher to the percentage, wont be changed
    :return:
    """
    R = random.Random(seed)
    occurrence_counter = Counter(y_train)
    Label_counts = sorted(occurrence_counter.items(), key=lambda x: x[1], reverse=True)
    largest_occur = Label_counts[0][1]
    min_num = int(largest_occur * reach_percent)
    new_x_train = list(x_train).copy()
    new_y_train = list(y_train).copy()
    for label, count in occurrence_counter.items():
        if count >= min_num:
            continue
        compensate_amount = min_num - count
        label_data = []
        for x, y in zip(x_train, y_train):
            if y == label:
                label_data.append(x)
        compensation = R.choices(label_data, k=compensate_amount)
        new_x_train.extend(compensation)
        new_y_train.extend([label] * compensate_amount)
    return new_x_train, new_y_train



"""
****************************************
 * @author: Xin Zhang
 * Date: 5/19/21
****************************************
"""


def find_top_n_faster(vlist, top_n, method="min", show_progress=True):
    if show_progress:
        print(f'finding top {top_n} values')

    top_n_values = vlist[:top_n]
    top_n_value_dict = dict((i, v) for i, v in enumerate(top_n_values))
    top_n_index_dict = dict((i, i) for i, v in enumerate(top_n_values))
    top_n_inverse_value = {v: k for k, v in top_n_value_dict.items()}
    if method == 'min':
        current_extreme_value = max(top_n_value_dict.values())
        current_extreme_key = top_n_inverse_value[current_extreme_value]
    elif method == 'max':
        current_extreme_value = min(top_n_value_dict.values())
        current_extreme_key = top_n_inverse_value[current_extreme_value]
    else:
        pass

    if show_progress:
        from tqdm import tqdm
        loop_pre_fun = tqdm
    else:
        loop_pre_fun = list
    for idx, value in enumerate(loop_pre_fun(vlist[top_n:])):
        list_index = idx + top_n
        if value < current_extreme_value and method == 'min':
            top_n_value_dict[current_extreme_key] = value
            top_n_index_dict[current_extreme_key] = list_index
            top_n_inverse_value = {v: k for k, v in top_n_value_dict.items()}
            current_extreme_value = max(top_n_inverse_value.keys())
            current_extreme_key = top_n_inverse_value[current_extreme_value]
        elif value > current_extreme_value and method == 'max':
            top_n_value_dict[current_extreme_key] = value
            top_n_index_dict[current_extreme_key] = list_index
            top_n_inverse_value = {v: k for k, v in top_n_value_dict.items()}
            current_extreme_value = min(top_n_inverse_value.keys())
            current_extreme_key = top_n_inverse_value[current_extreme_value]
        else:
            continue

    return list(top_n_value_dict.values()), list(top_n_index_dict.values())


def find_top_n(vlist, top_n, method="min", show_progress=True):
    if show_progress:
        print(f'finding top {top_n} values')
    top_n_l = vlist[:top_n]
    top_n_index = list(range(top_n))
    if show_progress:
        from tqdm import tqdm
        loop_pre_fun = tqdm
    else:
        loop_pre_fun = list

    if method == 'max':
        for idx, v in enumerate(loop_pre_fun(vlist[top_n:])):
            min_in_l = min(top_n_l)
            min_idx = top_n_l.index(min_in_l)
            if v > min_in_l:
                top_n_l[min_idx] = v
                top_n_index[min_idx] = idx + top_n

    elif method == 'min':
        for idx, v in enumerate(loop_pre_fun(vlist[top_n:])):
            max_in_l = max(top_n_l)
            max_idx = top_n_l.index(max_in_l)
            if v < max_in_l:
                top_n_l[max_idx] = v
                top_n_index[max_idx] = idx + top_n
    return top_n_l, top_n_index


def demo():
    """
    compare 2 algo to pick top n from
    @return:
    """
    import numpy as np
    import time
    a = list(np.random.uniform(low=-1.0, high=1.0, size=200000))
    top_n = 1000
    method = 'min'

    start = time.time()
    x1, y1 = find_top_n_faster(a, top_n, method, show_progress=True)
    print(x1, y1)
    print(time.time() - start)

    start = time.time()
    x2, y2 = find_top_n(a, top_n, method, show_progress=True)
    print(x2, y2)
    print(time.time() - start)

    print('verify------')
    print([a[i] for i in y1])
    print([a[i] for i in y2])
